#ifndef TESTSINGLECELLSINGLEMACHINEMODEL_HPP_
#define TESTSINGLECELLSINGLEMACHINEMODEL_HPP_

#include <cxxtest/TestSuite.h>

// Must be included before other cell_based headers
#include "CellBasedSimulationArchiver.hpp"

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>

#include <fstream>

#include "AbstractCellBasedTestSuite.hpp"
#include "SmartPointers.hpp"
#include "HoneycombMeshGenerator.hpp"
#include "WildTypeCellMutationState.hpp"
#include "DifferentiatedCellProliferativeType.hpp"
#include "UniformCellCycleModel.hpp"
#include "TypeSixMachineProperty.hpp"
#include "NodeBasedCellPopulation.hpp"
#include "GeneralisedLinearSpringForce.hpp"
#include "OffLatticeSimulation.hpp"
#include "TypeSixMachineStateAndPositionModifier.hpp"
#include "TypeSixSecretionEnumerations.hpp"
#include "UblasCustomFunctions.hpp"
#include "NodeBasedCellPopulationWithCapsules.hpp"
#include "ForwardEulerNumericalMethodForCapsules.hpp"
#include "GeneralisedCapsuleForce.hpp"

#include "TypeSixMachineCellKiller.hpp"

#include "MachineStateCountWriter.hpp"
#include "CapsuleOrientationWriter.hpp"
#include "CapsuleScalingWriter.hpp"
#include "CellIdWriter.hpp"
#include "CapsuleBasedDivisionRule.hpp"
#include "T6SSCellLabelWriter.hpp"
#include "MachineDataWriter.hpp"
#include "CellLabel.hpp"

#include "Debug.hpp"




// This test is always run sequentially (never in parallel)
#include "FakePetscSetup.hpp"

class TestSingleCellSingleMachineModel : public AbstractCellBasedTestSuite
{
public:
double mean_t6ss_cycle_time = 1.0/20.0; // Mean cycle time is 3 minutes.
double state_1_to_state_2_ratio = 2.7; // Observations 2.7 TssJ to TssB (check)
double first_simulation_stop_time = 3.0;
double second_simulation_stop_time = 7.0;
double min_cell_cycle_time=1.0;
double max_cell_cycle_time=1.29;
double timestep=1.0/1200.0;
double youngs_modulus=300.0;
double probability_target_cell=0.5;
double single_machine_k_1=0.0; 
double baseline_k_1=0.0; 
double baseline_k_2=0.0; 
double baseline_k_4=0.0; 
double baseline_k_5=30;
double baseline_k_3=baseline_k_5/state_1_to_state_2_ratio;
double baseline_k_6=0.0; 
double baseline_k_7=70.0; // From above calcs, assume K7 is big  
double baseline_machineMovementNoiseStrength=2.0; // Later define diffusion from literature membrane coef check paper microns squared per hour
unsigned num_realisations = 50u;
unsigned num_iterations=20;
double min_parameter=0.0;
double max_parameter=10.0;
unsigned initial_machine_state = 1u;
unsigned attacker_cell_label = 1u;
unsigned sample_rate = 1u;

std::vector<unsigned>  cell_number_at_archiving;
std::vector<unsigned>  cell_number_at_end_time;

std::string output_base_directory = "T6SSSingleCellGrowingMachineMovement"; 
std::string init_date = "InitialData";
std::string t6ss = "T6SS";
std::string parameter_label = "BaselineParameters";
std::string realisation_label = "Realisation";
std::string underscore = "_";
std::string fslash = "/";
std::string cells_in_layer_filename = "CompScoreAndParam.dat";
std::string output_directory_ic = output_base_directory + fslash + init_date; 
std::string output_filename_competitive_score_param = t6ss + underscore + parameter_label + underscore + cells_in_layer_filename;

    void TestSingleCellMachineMovementArchivingData()
    {
        EXIT_IF_PARALLEL;

        // Create a node
        std::vector<Node<3>*> nodes;
        nodes.push_back(new Node<3>(0u, Create_c_vector(4.0, 4.0,0.0)));
        
        // Connect nodes to mesh
        NodesOnlyMesh<3> mesh;
        mesh.ConstructNodesWithoutMesh(nodes, 150.5);

     
        // Add attributes to nodes
        mesh.GetNode(0u)->AddNodeAttribute(0.0);
        mesh.GetNode(0u)->rGetNodeAttributes().resize(NA_VEC_LENGTH);
        mesh.GetNode(0u)->rGetNodeAttributes()[NA_THETA] = 0.0 * M_PI;
        mesh.GetNode(0u)->rGetNodeAttributes()[NA_LENGTH] = 2.0;
        mesh.GetNode(0u)->rGetNodeAttributes()[NA_RADIUS] = 0.5;
        mesh.GetNode(0u)->rGetNodeAttributes()[NA_PHI] = M_PI/2.0;


        //Convert cells from nodes
        std::vector<CellPtr> cells;
        MAKE_PTR(WildTypeCellMutationState, p_state);
        MAKE_PTR(DifferentiatedCellProliferativeType, p_type);
        for (unsigned i=0; i<mesh.GetNumNodes(); i++)
        {
            UniformCellCycleModel* p_model = new UniformCellCycleModel();
            p_model->SetMinCellCycleDuration(min_cell_cycle_time);
            p_model->SetMaxCellCycleDuration(max_cell_cycle_time);
            CellPtr p_cell(new Cell(p_state, p_model));
            p_cell->SetCellProliferativeType(p_type);

            MAKE_PTR(TypeSixMachineProperty, p_property);
            p_cell->AddCellProperty(p_property);

            double birth_time = -0.8*RandomNumberGenerator::Instance()->ranf();
            p_cell->SetBirthTime(birth_time);			   
            
            double initial_length=p_property->GetCellLengthAfterDivision();
            double cell_age  = p_cell->GetAge();//+SimulationTime::Instance()->GetTimeStep();
            double cell_cycle_time =  p_model->GetCellCycleDuration() ;
            double division_length = 2.0*(initial_length + 1.0*mesh.GetNode(i)->rGetNodeAttributes()[NA_RADIUS]);
            double new_length = initial_length + (division_length - initial_length)*cell_age/cell_cycle_time;
            mesh.GetNode(i)->rGetNodeAttributes()[NA_LENGTH] = new_length;
            cells.push_back(p_cell);
        }



        NodeBasedCellPopulationWithCapsules<3> cell_population(mesh, cells);
        cell_population.AddCellWriter<CellIdWriter>();
        cell_population.AddCellWriter<CapsuleOrientationWriter>();
        cell_population.AddCellWriter<CapsuleScalingWriter>();
        cell_population.AddCellWriter<T6SSCellLabelWriter>();
        cell_population.AddCellWriter<MachineStateCountWriter>();
        cell_population.AddCellWriter<MachineDataWriter>();

        boost::shared_ptr<AbstractCentreBasedDivisionRule<3,3> > p_division_rule(new CapsuleBasedDivisionRule<3,3>());
        cell_population.SetCentreBasedDivisionRule(p_division_rule);
        
        // Create a simulation
        OffLatticeSimulation<3> simulator(cell_population);
        
        //Set output directory
        simulator.SetOutputDirectory(output_directory_ic);
        simulator.SetDt(timestep);
        simulator.SetSamplingTimestepMultiple(sample_rate);
        
        // define numerical method class
        auto p_numerical_method = boost::make_shared<ForwardEulerNumericalMethodForCapsules<3,3>>();
        simulator.SetNumericalMethod(p_numerical_method);

        // Create a force law and pass it to the simulation
        auto p_capsule_force = boost::make_shared<GeneralisedCapsuleForce<3>>();
        p_capsule_force->SetYoungModulus(youngs_modulus);
        p_capsule_force->SetSolveMechanicsIn3D(false); // make the cells stay in the plane by setting this to false.
        simulator.AddForce(p_capsule_force);

        simulator.SetEndTime(first_simulation_stop_time); 
        simulator.Solve();
        TS_ASSERT_EQUALS(simulator.rGetCellPopulation().GetNumRealCells(),1u);
        MAKE_PTR_ARGS(TypeSixMachineCellKiller<3>, p_killer, (&cell_population));
        simulator.AddCellKiller(p_killer);
    
    
        unsigned machine_id=1u;

        NodeBasedCellPopulationWithCapsules<3>& rcapsule_pop=(static_cast<NodeBasedCellPopulationWithCapsules<3>&>(simulator.rGetCellPopulation()));
        for (NodeBasedCellPopulationWithCapsules<3>::Iterator cell_iter = rcapsule_pop.Begin();
                        cell_iter != rcapsule_pop.End(); ++cell_iter)
        {
            // Get T6SS property
            CellPropertyCollection collection = cell_iter->rGetCellPropertyCollection().template GetProperties<TypeSixMachineProperty>();
            boost::shared_ptr<TypeSixMachineProperty> p_property = boost::static_pointer_cast<TypeSixMachineProperty>(collection.GetProperty());

            // Add a machine to the cell
            Node<3>* p_node = rcapsule_pop.GetNodeCorrespondingToCell(*cell_iter);
            
            // Set initial phi angle of macine on cell
            double T6SS_phi=(RandomNumberGenerator::Instance()->ranf()-0.5)*2*M_PI;

            // Set (x,y,z) positions
            double T6_ss_x_coord = 1.0*(p_node->rGetNodeAttributes()[NA_LENGTH])*(RandomNumberGenerator::Instance()->ranf()-0.5);
            double T6_ss_y_coord =  cos(T6SS_phi)*(p_node->rGetNodeAttributes()[NA_RADIUS]);
            double T6_ss_z_coord = sin(T6SS_phi)*(p_node->rGetNodeAttributes()[NA_RADIUS]);
            
            // Initialise machine
            p_property->InitialiseMachine(T6_ss_x_coord,T6_ss_y_coord,T6_ss_z_coord,initial_machine_state,machine_id);
            
            // increment id
            machine_id++; 

            // Set attacker cell label (1)
            p_property->SetCellTypeLabel(attacker_cell_label);
            
        }
        
        
        // Save simulator details to file?
        CellBasedSimulationArchiver<3, OffLatticeSimulation<3> >::Save(&simulator);


        // Free up memory?
        SimulationTime::Destroy();
        SimulationTime::Instance()->SetStartTime(first_simulation_stop_time);
        RandomNumberGenerator::Destroy();
    }
    
    
    void TestSingleCellMachineMovement()
    {

        // Set results handlers
        OutputFileHandler results_handler(output_directory_ic + "/", false); //Create output file handler
        out_stream cells_in_layer_file = results_handler.OpenOutputFile(output_filename_competitive_score_param);
            
        OffLatticeSimulation<3>* p_simulator;

        p_simulator = CellBasedSimulationArchiver<3, OffLatticeSimulation<3> >::Load(output_directory_ic, first_simulation_stop_time);


        // set all machine properties to have parameter
        NodeBasedCellPopulationWithCapsules<3>& rcapsule_pop=(static_cast<NodeBasedCellPopulationWithCapsules<3>&>(p_simulator->rGetCellPopulation()));
        for (NodeBasedCellPopulationWithCapsules<3>::Iterator cell_iter = rcapsule_pop.Begin();
                        cell_iter != rcapsule_pop.End(); ++cell_iter)
        {
            // Get T6SS property
            CellPropertyCollection collection = cell_iter->rGetCellPropertyCollection().template GetProperties<TypeSixMachineProperty>();
            boost::shared_ptr<TypeSixMachineProperty> p_property = boost::static_pointer_cast<TypeSixMachineProperty>(collection.GetProperty());
            
            unsigned label= p_property->GetCellTypeLabel();
            if(label==attacker_cell_label)
            {

                
                p_property->SetMachineParameters(baseline_k_1, baseline_k_2, baseline_k_3, baseline_k_4, baseline_k_5, baseline_k_6, baseline_k_7,baseline_machineMovementNoiseStrength);         

            }
        }
        

        // Set simulator output directory
        std::stringstream out;
        out << t6ss << underscore << parameter_label; 
        std::string output_directory = output_base_directory + fslash + out.str();

        p_simulator->SetOutputDirectory(output_directory);

        // Create a T6SSS modifier and pass it to the simulation
        MAKE_PTR(TypeSixMachineStateAndPositionModifier<3>, p_modifier);
        p_modifier->SetOutputDirectory(output_directory);
        p_simulator->AddSimulationModifier(p_modifier);
        // Add cell killer

        cell_number_at_archiving=rcapsule_pop.ComputeNumberOfCellsWithGivenLabel();

        p_simulator->SetEndTime(second_simulation_stop_time); 
        TS_ASSERT_THROWS_NOTHING(p_simulator->Solve());

        // Compute competition score
        cell_number_at_end_time=rcapsule_pop.ComputeNumberOfCellsWithGivenLabel();
        double pop_1_term=(double)cell_number_at_end_time[1]/(double)cell_number_at_archiving[1];
        double pop_0_term=(double)cell_number_at_end_time[0]/(double)cell_number_at_archiving[0];
        
        double compeition_metric=1.0/(second_simulation_stop_time - first_simulation_stop_time)*log(pop_1_term/pop_0_term);

        *cells_in_layer_file  << "\t" << compeition_metric << "\n";


        //Tidying up
        SimulationTime::Instance()->Destroy();
        SimulationTime::Instance()->SetStartTime(first_simulation_stop_time);

        // compute competion score on the fly so we get a feel for data without having to go to file
        delete p_simulator;
        cells_in_layer_file->close();
    }

    void NoTestSingleCellMachineMovementMultipleRealisations()
    {
        //Create output file handler
        OutputFileHandler results_handler(output_directory_ic + fslash, false); 
        out_stream cells_in_layer_file = results_handler.OpenOutputFile(output_filename_competitive_score_param);
        

        for (unsigned realisation_index=1u;realisation_index<num_realisations;realisation_index++)
        {
            // Add simulation and lattice together
            OffLatticeSimulation<3>* p_simulator;
            p_simulator = CellBasedSimulationArchiver<3, OffLatticeSimulation<3> >::Load(output_directory_ic, first_simulation_stop_time);
            
            // Ensure each realisation is using a different random seed
            RandomNumberGenerator::Instance()->Reseed(100*realisation_index);


            // set all machine properties to have parameter
            NodeBasedCellPopulationWithCapsules<3>& rcapsule_pop=(static_cast<NodeBasedCellPopulationWithCapsules<3>&>(p_simulator->rGetCellPopulation()));
            for (NodeBasedCellPopulationWithCapsules<3>::Iterator cell_iter = rcapsule_pop.Begin();
                            cell_iter != rcapsule_pop.End(); ++cell_iter)
            {
                // Get T6SS property
                CellPropertyCollection collection = cell_iter->rGetCellPropertyCollection().template GetProperties<TypeSixMachineProperty>();
                boost::shared_ptr<TypeSixMachineProperty> p_property = boost::static_pointer_cast<TypeSixMachineProperty>(collection.GetProperty());
                unsigned label= p_property->GetCellTypeLabel();
                
                if(label==attacker_cell_label)
                {

                    p_property->SetMachineParameters(single_machine_k_1, baseline_k_2, baseline_k_3, baseline_k_4, baseline_k_5, baseline_k_6, baseline_k_7,baseline_machineMovementNoiseStrength);         

                }
            }
            
            // Generate string of realisation index with leading zero 
            std::string realisation_string = std::to_string(realisation_index);
            //std::string new_realisation_index = (realisation_string.length() < 2) ? "00" + realisation_string : realisation_string;
             std::string new_realisation_index;
             
             
             if (realisation_string.length() < 2) 
             {
                new_realisation_index = "000" + realisation_string;
             } 
             else if (realisation_string.length() < 3) 
             {
                new_realisation_index = "00" + realisation_string;
             } 
             else 
             {
                new_realisation_index = realisation_string;
             }

            //Set output directory
            std::stringstream out;
            out << realisation_label << underscore << new_realisation_index; 
            std::string output_directory = output_base_directory + fslash + parameter_label + fslash +  out.str();


            p_simulator->SetOutputDirectory(output_directory);



            p_simulator->SetOutputDirectory(output_directory);

            // Create a T6SSS modifier and pass it to the simulation
            MAKE_PTR(TypeSixMachineStateAndPositionModifier<3>, p_modifier);
            p_modifier->SetOutputDirectory(output_directory);
            p_simulator->AddSimulationModifier(p_modifier);
            
            
            // Compute number of cell with given labels
            cell_number_at_archiving=rcapsule_pop.ComputeNumberOfCellsWithGivenLabel();


            // Set simulation time to end
            p_simulator->SetEndTime(second_simulation_stop_time); 
            TS_ASSERT_THROWS_NOTHING(p_simulator->Solve());


            // Free up memory
            SimulationTime::Instance()->Destroy();
            SimulationTime::Instance()->SetStartTime(first_simulation_stop_time);


            // compute competion score on the fly so we get a feel for data without having to go to file
            delete p_simulator;
        }
    }



};

#endif /*TESTSINGLECELLSINGLEMACHINEMODEL_HPP_*/
